-----------------------------------------------------------------------------------------------
-- Client Lua Script for AuctionScanner
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- AuctionScanner Module Definition
-----------------------------------------------------------------------------------------------
local AuctionScanner = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local knMaxItemID = 99999		-- The maximum item ID to check while scanning. May need to be increased in the future as content patches/expansions add more items.

-----------------------------------------------------------------------------------------------
-- Variables
-----------------------------------------------------------------------------------------------
local iScanningIndex = nil
local tItems = {}
local oTimer = nil
local bProcessing = false		-- True if we're processing an item, used to prevent unlikely case of timer firing while still processing last call.
local bScanning = false			-- Only true if we initiated a scan, used to minimise pollution from normal AH usage.

local tAuctionValues = {}		-- Table keyed by item ID.
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AuctionScanner:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	self.tAuctionValues = self.tAuctionValues or {}

    return o
end

function AuctionScanner:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- AuctionScanner OnLoad
-----------------------------------------------------------------------------------------------
function AuctionScanner:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("AuctionScanner.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)

    Apollo.GetPackage("Gemini:Hook-1.0").tPackage:Embed(self)
end

-----------------------------------------------------------------------------------------------
-- AuctionScanner OnDocLoaded
-----------------------------------------------------------------------------------------------
function AuctionScanner:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterEventHandler("ItemAuctionSearchResults", "OnItemAuctionSearchResults", self)

		-- Do additional Addon initialization here
	    self.MarketplaceAuction = Apollo.GetAddon("MarketplaceAuction")

		-- Hook the tooltip window
	    self:PostHook(Apollo.GetAddon("ToolTips"), "CreateCallNames")

		-- Hook the auction window
		if self.MarketplaceAuction then
		    self:PostHook(self.MarketplaceAuction, "Initialize")
			self:PostHook(self.MarketplaceAuction, "OnDestroy")
			self:PostHook(self.MarketplaceAuction, "InitializeSell")
			self:PostHook(self.MarketplaceAuction, "OnSellListItemCheck")
		end
		self.iItemsProcessed = 0
	end
end

-----------------------------------------------------------------------------------------------
-- AuctionScanner Functions
-----------------------------------------------------------------------------------------------
function AuctionScanner:OnAuctionScannerOn()
	self:ScanAuctions()
end

function AuctionScanner:Initialize(luaCaller)
	if luaCaller == self.MarketplaceAuction then
		self.MarketplaceAuction = luaCaller
		-- Add the button, destroying first if it already exists
        if self.ScanButton ~= nil then
			self.ScanButton:Destroy()
		end
        self.ScanButton = Apollo.LoadForm(self.xmlDoc, "ScanButton", luaCaller.wndMain, self)
        self.AuctionBlocker = Apollo.LoadForm(self.xmlDoc, "AuctionBlocker", luaCaller.wndMain, self)
		self.AuctionBlocker:Show(false)
	end
end

function AuctionScanner:OnDestroy(luaCaller)
	if luaCaller == self.MarketplaceAuction then
		if self.bScanning then
			Print("AuctionScanner: Auction window closed, ending scan.")
			self:EndScan()
		end
	end
end

-- Adds the stack count to the list of sellable items.
function AuctionScanner:InitializeSell()
	local wndContainer = self.MarketplaceAuction.wndMain:FindChild("SellLeftSideItemList")
	if wndContainer then
		local tChildren = wndContainer:GetChildren()
		for iIndex, tChild in pairs(tChildren) do
			local wndIcon = tChild:FindChild("ListItemIcon")
			if wndIcon then
				local oItem = tChild:GetData()
				if oItem then
					wndIcon:SetText(oItem:GetStackCount())
					wndIcon:SetFont("CRB_Pixel_O")
				end
			end
		end
	end
end

-- Adds the stack count to the big icon on the right-hand area while selling.
function AuctionScanner:OnSellListItemCheck(wndHandler, wndControl)
	local wndButton = self.MarketplaceAuction.wndMain:FindChild("CreateSellOrderBtn")
	if wndButton then
		local oItem = wndButton:GetData()
		if oItem then
			local wndBigIcon = self.MarketplaceAuction.wndMain:FindChild("SellContainer"):FindChild("SellRightSide"):FindChild("BigIcon")
			if wndBigIcon then
				wndBigIcon:SetText(oItem:GetStackCount())
				wndBigIcon:SetFont("CRB_Pixel_O")
			end
		end
	end
end


function AuctionScanner:ScanAuctions()
	-- Build a list of all items flagged auctionable.
	local tItems = {}
	local index = 1
	for index = 1, knMaxItemID do
		local oItem = Item.GetDataFromId(index)
		if oItem ~= nil then
			if oItem:IsAuctionable() then
				table.insert(tItems, oItem:GetItemId())		-- Store the item ID in a table.
			end
		end
	end
		
	Print("AuctionScanner: Found " .. #tItems .. " auctionable items.")
	
	self.tItems = tItems
	self.iScanningIndex = 0

	self.oTimer = ApolloTimer.Create(0.0, false, "OnScanAuctionsTimerTick", self)	
end

-- Used to stop the default auction addon from noticing the search results and trying to render them.
function AuctionScanner:DummyOnItemAuctionSearchResults(nPage, nTotalResults, tAuctions)
end

-- Sends the requests for each item. Note this performs around 10500 calls to the server, not sure whether CRB would be ok with this but had no response as yet.
function AuctionScanner:OnScanAuctionsTimerTick()
	if self.bProcessing == true then
		return
	end
	self.bProcessing = true
	self.iItemsProcessed = 0
	
	for iIndex = 1, #self.tItems do
		self.iScanningIndex = iIndex
	
		if self.iScanningIndex > #self.tItems then
		else
			local nItemId = self.tItems[self.iScanningIndex]
			-- Gets a list of auctions for the specific item, with the lowest buyout first.
			MarketplaceLib.RequestItemAuctionsByItems({ nItemId }, 0, MarketplaceLib.AuctionSort.Buyout, false, nil, nil, nil, nil)
		end
	end
end


function AuctionScanner:OnItemAuctionSearchResults(nPage, nTotalResults, tAuctions)
	if self.bScanning == true and self.tItems then
		self.iItemsProcessed = self.iItemsProcessed + 1
		self.ScanButton:SetText("Scanning: " .. math.floor(self.iItemsProcessed / #self.tItems * 100) .. "%")
		
		if self.iItemsProcessed >= #self.tItems then
			self:EndScan()
			Print("AuctionScanner: All items processed, ending scan.")
	
			self.bProcessing = false
		end
	end
		
	if #tAuctions > 0 and self.bScanning then
		-- Since we sorted by lowest buyout first, we're only interested in the first result on the first page of results.
		-- This is implemented as a loop even though we usually want the first item, as sometimes the earliest items are
		-- those with no buyout and only a bid price. As such, we iterate through until we find one with a buyout, or hit
		-- the end of the first page, in which case we give up.
		for iAuctionIndex, aucCurr in pairs(tAuctions) do
			local itemCurr = aucCurr:GetItem()
			local nItemId = itemCurr:GetItemId()
			local bIsOwnAuction = aucCurr:IsOwned()
			local nBuyoutPrice = aucCurr:GetBuyoutPrice():GetAmount()
			local nDefaultBid = math.max(aucCurr:GetMinBid():GetAmount(), aucCurr:GetCurrentBid():GetAmount())
			
			if nBuyoutPrice > 0 then
				-- Found an item with a buyout price, store it and exit the function.
				local tItemValues = {}
				tItemValues.nBuyoutPrice = nBuyoutPrice
				tItemValues.nId = nItemId
				tItemValues.iNumberOfAuctions = #tAuctions
				tItemValues.dtLastUpdate = os.date()
				
				self.tAuctionValues[nItemId] = tItemValues
				return
			end
		end
	end
end

-- Adds buyout price information to the tooltip.
function AuctionScanner:CreateCallNames(luaCaller)
	-- Essentially ripped this from CommodityStats, credit goes to Boe2.
    local origItemTooltip = Tooltip.GetItemTooltipForm
    Tooltip.GetItemTooltipForm = function(luaCaller, wndControl, item, bStuff, nCount)
        wndControl:SetTooltipDoc(nil)
        local wndTooltip, wndTooltipComp = origItemTooltip(luaCaller, wndControl, item, bStuff, nCount)
        if (wndTooltip ~= nil) and item:IsAuctionable() then
            local itemID = item:GetItemId()
            local itemValue = self:GetItemValue(itemID)
            if itemValue then
                local offset = 20
                local extra = Apollo.LoadForm(self.xmlDoc, "ItemTooltipAdditions", wndTooltip, self)
                extra:FindChild("BuyoutPrice"):SetAmount(itemValue.nBuyoutPrice)

				extra:FindChild("LastUpdated"):SetText("Last seen: " .. itemValue.dtLastUpdate)

                local eLeft, eTop, eRight, eBottom = extra:GetAnchorOffsets()
                local nLeft, nTop, nRight, nBottom = wndTooltip:GetAnchorOffsets()
                extra:SetAnchorOffsets(nLeft, nBottom - offset, nRight, nBottom + (eBottom - eTop - offset))
                wndTooltip:SetAnchorOffsets(nLeft, nTop, nRight, nBottom + (eBottom - eTop - offset))
            end
        end
        return wndTooltip, wndTooltipComp
    end
end


function AuctionScanner:GetItemValue(nItemId)
	return self.tAuctionValues[nItemId]
end

-----------------------------------------------------------------------------------------------
-- AuctionScannerForm Functions
-----------------------------------------------------------------------------------------------
-- when the OK button is clicked
function AuctionScanner:OnOK()
	self.wndMain:Close() -- hide the window
end

-- when the Cancel button is clicked
function AuctionScanner:OnCancel()
	self.wndMain:Close() -- hide the window
end

function AuctionScanner:OnScanButton()
	if self.bScanning then
		-- Can no longer abort scans as the results still come in (and as the AH handler is restored, slows the game to a crawl for a while).
		--self:EndScan()
	else
		self:StartScan()
	end
end


function AuctionScanner:StartScan()
	self.ScanButton:SetText("Scanning: 0%")
	self.ScanButton:Enable(false)
	self.AuctionBlocker:Show(true)	
	self.AuctionBlocker:ToFront()
	self.bScanning = true
	self.OriginalOnItemAuctionSearchResults = self.MarketplaceAuction.OnItemAuctionSearchResults
	self.MarketplaceAuction.OnItemAuctionSearchResults = self.DummyOnItemAuctionSearchResults 
	self:OnAuctionScannerOn()
end

function AuctionScanner:EndScan()
	if self.oTimer then
		self.oTimer:Stop()
		self.oTimer = nil
	end
	self.bProcessing = false
	self.bScanning = false
	self.MarketplaceAuction.OnItemAuctionSearchResults = self.OriginalOnItemAuctionSearchResults 
	self.AuctionBlocker:Show(false)	
	self.ScanButton:SetText("Scan Auctions")
	self.ScanButton:SetTooltip("Scan all active auctions and store the cheapest buyout value for each. The auction window will be unavailable during the scan.")
	self.ScanButton:Enable(true)
end


-----------------------------------------------------------------------------------------------
-- AuctionScanner Instance
-----------------------------------------------------------------------------------------------
local AuctionScannerInst = AuctionScanner:new()
AuctionScannerInst:Init()


-----------------------------------------------------------------------------------------------
-- AuctionScanner Persistence
-----------------------------------------------------------------------------------------------
function AuctionScanner:OnSave(eLevel)
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Realm then
	    local save = 
		{
			tAuctionValues = self.tAuctionValues 
	    }
    	return save
    end 

    return nil
end

function AuctionScanner:OnRestore(eLevel, tData)
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Realm then
	    self.tAuctionValues = tData.tAuctionValues or {}
    end 
end
